@@ -1,4 +1,5 @@
 """Mastertherm API Client Class, handle integration."""
+import asyncio
 import logging
 import time
 
@@ -7,6 +8,8 @@ from hashlib import sha1
 from json.decoder import JSONDecodeError
 from urllib.parse import urljoin, quote_plus
 
+from natsort import natsorted
+
 from aiohttp import ClientSession, ClientConnectionError, ContentTypeError
 
 from .const import (
@@ -34,6 +37,8 @@ from .exceptions import (
     MasterthermUnsupportedVersion,
     MasterthermTokenInvalid,
     MasterthermResponseFormatError,
+    MasterthermPumpError,
+    MasterthermServerTimeoutError,
 )
 
 _LOGGER: logging.Logger = logging.getLogger(__name__)
@@ -100,15 +105,12 @@ class MasterthermAPI:
             if self.__expires <= datetime.now():
                 return True
 
-        if self.__expires <= datetime.now():
-            return True
-
         return False
 
     async def __post(self, url: str, params: str) -> dict:
         """Push updates to the API."""
         if await self.__token_expired():
-            await self.connect()
+            await self.__connect_refresh()
 
         try:
             if self.__api_version == "v1":
@@ -131,54 +133,52 @@ class MasterthermAPI:
                         "Connection": "close",
                     },
                 )
+
+            response_json = await response.json()
         except ClientConnectionError as ex:
             _LOGGER.error("Client Connection Error: %s", ex)
             raise MasterthermConnectionError("3", "Client Connection Error") from ex
         except JSONDecodeError as ex:
-            _LOGGER.error("JSON Decode Error: %s", ex)
-            raise MasterthermConnectionError("3", "JSON Decode Error") from ex
+            response_text = await response.text()
+            _LOGGER.error(
+                "JSON Decode Error: %s:%s",
+                response.status,
+                response_text,
+            )
+            raise MasterthermResponseFormatError(response.status, response_text) from ex
+        except ContentTypeError as ex:
+            if response.status == 504:
+                raise MasterthermServerTimeoutError(
+                    response.status, response.reason
+                ) from ex
+
+            response_text = await response.text()
+            if response_text == "User not logged in":
+                raise MasterthermTokenInvalid(response.status, response_text) from ex
+            else:
+                _LOGGER.error(
+                    "Mastertherm API some other error: %s:%s",
+                    response.status,
+                    response_text,
+                )
+                raise MasterthermConnectionError(response.status, response_text) from ex
 
         # Version 2 responds with an error and json.
         # We should only get something other than 200 if the servers are down.
         if response.status != 200:
-            try:
-                response_json = await response.json()
-
-                # Deal with the v2 error if we have a response
-                if response_json["status"]["id"] == 401:
-                    raise MasterthermTokenInvalid("1", response_json)
-                else:
-                    _LOGGER.error("Mastertherm API some other error: %s", response_json)
-                    raise MasterthermResponseFormatError("2", response_json)
-            except ContentTypeError as ex:
-                response_text = await response.text()
-                _LOGGER.error(
-                    "Mastertherm API Connection Error %s:%s",
-                    str(response.status),
-                    response_text,
-                )
-                raise MasterthermConnectionError(
-                    str(response.status), response_text
-                ) from ex
-
-        # Convert to JSON if possible and return the result
-        try:
-            response_json = await response.json()
-        except ContentTypeError as exc:
-            response_text = await response.text()
-            if response_text == "User not logged in":
-
-                raise MasterthermTokenInvalid("1", response_text) from exc
+            # Deal with the v2 error if we have a response
+            if response_json["status"]["id"] == 401:
+                raise MasterthermTokenInvalid(response.status, response_json)
             else:
-                _LOGGER.error("Mastertherm API some other error: %s", response_text)
-                raise MasterthermResponseFormatError("2", response_text) from exc
+                _LOGGER.error("Mastertherm API some other error: %s", response_json)
+                raise MasterthermResponseFormatError(response.status, response_json)
 
         return response_json
 
     async def __get(self, url: str, params: str) -> dict:
         """Get updates from the API, for old this mostly uses Post."""
         if await self.__token_expired():
-            await self.connect()
+            await self.__connect_refresh()
 
         try:
             if self.__api_version == "v1":
@@ -200,70 +200,49 @@ class MasterthermAPI:
                         "Connection": "close",
                     },
                 )
+
+            response_json = await response.json()
         except ClientConnectionError as ex:
             _LOGGER.error("Client Connection Error: %s", ex)
             raise MasterthermConnectionError("3", "Client Connection Error") from ex
         except JSONDecodeError as ex:
-            _LOGGER.error("JSON Decode Error: %s", ex)
-            raise MasterthermConnectionError("3", "JSON Decode Error") from ex
-
-        # Version 2 responds with an error and json.
-        # We should only get something other than 200 if the servers are down.
-        if response.status != 200:
-            try:
-                response_json = await response.json()
-
-                # Deal with the v2 error if we have a response
-                if response_json["status"]["id"] == 401:
-                    raise MasterthermTokenInvalid("1", response_json)
-                else:
-                    _LOGGER.error("Mastertherm API some other error: %s", response_json)
-                    raise MasterthermResponseFormatError("2", response_json)
-            except ContentTypeError as ex:
-                response_text = await response.text()
-                _LOGGER.error(
-                    "Mastertherm API Connection Error %s:%s",
-                    str(response.status),
-                    response_text,
-                )
-                raise MasterthermConnectionError(
-                    str(response.status), response_text
+            response_text = await response.text()
+            _LOGGER.error(
+                "JSON Decode Error: %s:%s",
+                response.status,
+                response_text,
+            )
+            raise MasterthermResponseFormatError(response.status, response_text) from ex
+        except ContentTypeError as ex:
+            if response.status == 504:
+                raise MasterthermServerTimeoutError(
+                    response.status, response.reason
                 ) from ex
 
-        # Convert to JSON if possible and return the result
-        try:
-            response_json = await response.json()
-        except ContentTypeError as exc:
             response_text = await response.text()
             if response_text == "User not logged in":
-
-                raise MasterthermTokenInvalid("1", response_text) from exc
+                raise MasterthermTokenInvalid(response.status, response_text) from ex
             else:
-                _LOGGER.error("Mastertherm API some other error: %s", response_text)
-                raise MasterthermResponseFormatError("2", response_text) from exc
+                _LOGGER.error(
+                    "Mastertherm API some other error: %s:%s",
+                    response.status,
+                    response_text,
+                )
+                raise MasterthermConnectionError(response.status, response_text) from ex
+
+        # Version 2 responds with an error and json.
+        if response.status != 200:
+            # Deal with the v2 error if we have a response
+            if response_json["status"]["id"] == 401:
+                raise MasterthermTokenInvalid(response.status, response_json)
+            else:
+                _LOGGER.error("Mastertherm API some other error: %s", response_json)
+                raise MasterthermResponseFormatError(response.status, response_json)
 
         return response_json
 
-    def get_url(self) -> str:
-        """Return the API URL Used.
-
-        Returns:
-            URL(str): The API URL for the version."""
-        if self.__api_version == "v1":
-            return URL_BASE
-        else:
-            return URL_BASE_NEW
-
-    async def connect(self) -> dict:
-        """Perform the connection to the Mastertherm API Server:
-
-        Returns:
-             devices (dict): Return the list of devices, modules and units.
-
-        Raises:
-            MasterthermConnectionError - Failed to Connect
-            MasterthermAuthenticationError - Failed to Authenticate
-            MasterthermUnsupportedRole - Role is not in supported roles"""
+    async def __connect_refresh(self) -> dict:
+        """Perform the Connect only or refresh token."""
         # Connect based on requirements
         if self.__api_version == "v1":
             # Clear out cookies, clears the auth token.
@@ -320,18 +299,32 @@ class MasterthermAPI:
                 seconds=response_json["expires_in"]
             )
 
+        return response_json
+
+    def get_url(self) -> str:
+        """Return the API URL Used.
+
+        Returns:
+            URL(str): The API URL for the version."""
+        if self.__api_version == "v1":
+            return URL_BASE
+        else:
+            return URL_BASE_NEW
+
+    async def connect(self) -> dict:
+        """Perform the connection to the Mastertherm API Server:
+
+        Returns:
+             devices (dict): Return the list of devices, modules and units.
+
+        Raises:
+            MasterthermConnectionError - Failed to Connect
+            MasterthermAuthenticationError - Failed to Authenticate
+            MasterthermUnsupportedRole - Role is not in supported roles"""
+        response_json = await self.__connect_refresh()
+        if self.__api_version == "v2":
             # Get the Modules as this now has moved to outside of the auth process
-            url = urljoin(URL_BASE_NEW, URL_MODULES_NEW)
-            response = await self.__session.get(
-                url, headers={"Authorization": f"Bearer {self.__token}"}
-            )
-
-            # Response shoudl always be 200 even for login failures
-            if response.status != 200:
-                error_msg = await response.text()
-                raise MasterthermConnectionError(str(response.status), error_msg)
-
-            response_json = await response.json()
+            response_json = await self.__get(url=URL_MODULES_NEW, params=None)
 
         # Next is same for old and new process, doing a double check just incase
         if response_json["returncode"] != 0:
@@ -360,7 +353,9 @@ class MasterthermAPI:
         Raises:
             MasterthermConnectionError - General Connection Issue
             MasterthermTokenInvalid - Token has expired or is invalid
-            MasterthermResponseFormatError - Some other issue, probably temporary"""
+            MasterthermResponseFormatError - Some other issue, probably temporary
+            MasterthermServerTimeoutError - Server Timed Out more than once."""
+        retry = False
         params = f"moduleid={module_id}&unitid={unit_id}&application=android"
 
         try:
@@ -368,8 +363,16 @@ class MasterthermAPI:
                 url=URL_PUMPINFO if self.__api_version == "v1" else URL_PUMPINFO_NEW,
                 params=params,
             )
-        except MasterthermTokenInvalid:
+        except MasterthermTokenInvalid as ex:
+            _LOGGER.warning("Token Expired Early Retry: %s:%s", ex.status, ex.message)
+            retry = True
             self.__expires = None
+        except MasterthermServerTimeoutError as ex:
+            _LOGGER.warning("API Timed Out Retry: %s:%s", ex.status, ex.message)
+            retry = True
+
+        if retry:
+            await asyncio.sleep(0.5)
             response_json = await self.__get(
                 url=URL_PUMPINFO if self.__api_version == "v1" else URL_PUMPINFO_NEW,
                 params=params,
@@ -393,7 +396,10 @@ class MasterthermAPI:
         Raises:
             MasterthermConnectionError - General Connection Issue
             MasterthermTokenInvalid - Token has expired or is invalid
-            MasterthermResponseFormatError - Some other issue, probably temporary"""
+            MasterthermResponseFormatError - Some other issue, probably temporary
+            MasterthermPumpDisconnected - Pump is unavailable, disconnected or offline.
+            MasterthermServerTimeoutError - Server Timed Out more than once."""
+        retry = False
         params = f"moduleId={module_id}&deviceId={unit_id}&application=android&"
         if last_update_time is None:
             params = (
@@ -411,13 +417,29 @@ class MasterthermAPI:
                 url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
                 params=params,
             )
-        except MasterthermTokenInvalid:
+        except MasterthermTokenInvalid as ex:
+            _LOGGER.warning(
+                "Token Expired Early Retrying: %s:%s", ex.status, ex.message
+            )
+            retry = True
             self.__expires = None
+        except MasterthermServerTimeoutError as ex:
+            _LOGGER.warning("API Timed Out Retry: %s:%s", ex.status, ex.message)
+            retry = True
+
+        if retry:
+            await asyncio.sleep(0.5)
             response_json = await self.__get(
                 url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
                 params=params,
             )
 
+        # Check for Errors with the Pump.
+        error_id = response_json["error"]["errorId"]
+        if error_id != 0:
+            raise MasterthermPumpError(error_id, response_json["error"]["errorMessage"])
+
+        # No error process the response.
         if response_json["data"] != {}:
             data_file = ""
             if self.__api_version == "v1":
@@ -431,6 +453,13 @@ class MasterthermAPI:
             response_json["data"]["varData"] = response_json["data"][data_file]
             del response_json["data"][data_file]
 
+            # Sort the Data
+            sorted_reg = {}
+            device_reg = response_json["data"]["varData"][str(unit_id).zfill(3)]
+            for key in natsorted(device_reg.keys()):
+                sorted_reg[key] = device_reg[key]
+            response_json["data"]["varData"][str(unit_id).zfill(3)] = sorted_reg
+
         return response_json
 
     async def set_device_data(
@@ -452,7 +481,9 @@ class MasterthermAPI:
         Raises:
             MasterthermConnectionError - General Connection Issue
             MasterthermTokenInvalid - Token has expired or is invalid
-            MasterthermResponseFormatError - Some other issue, probably temporary"""
+            MasterthermResponseFormatError - Some other issue, probably temporary
+            MasterthermServerTimeoutError - Server Timed Out more than once."""
+        retry = False
         params = (
             f"moduleId={module_id}&deviceId={unit_id}&"
             + "configFile=varfile_mt1_config&messageId=1&errorResponse=true&"
@@ -466,13 +497,36 @@ class MasterthermAPI:
                 else URL_POSTUPDATE_NEW,
                 params=params,
             )
-        except MasterthermTokenInvalid:
+        except MasterthermTokenInvalid as ex:
+            _LOGGER.warning(
+                "Token Expired Early Retrying: %s:%s", ex.status, ex.message
+            )
+            retry = True
             self.__expires = None
-            response_json = await self.__get(
-                url=URL_PUMPDATA if self.__api_version == "v1" else URL_PUMPDATA_NEW,
+        except MasterthermServerTimeoutError as ex:
+            _LOGGER.warning("API Timed Out Retry: %s:%s", ex.status, ex.message)
+            retry = True
+
+        if retry:
+            await asyncio.sleep(0.5)
+            response_json = await self.__post(
+                url=URL_POSTUPDATE
+                if self.__api_version == "v1"
+                else URL_POSTUPDATE_NEW,
                 params=params,
             )
 
+        _LOGGER.debug("Set Response Received: %s", response_json)
+
+        # Check the response for any errors:
+        if response_json["error"]["errorId"] != 0:
+            _LOGGER.error(
+                "Client Connection Error: %s, Message: %s",
+                response_json["error"]["errorId"],
+                response_json["error"]["errorMessage"],
+            )
+            return False
+
         # Get the Value that is returned and set.
         if self.__api_version == "v1":
             data_file = ""
