@@ -8,11 +8,14 @@ from aiohttp import ClientSession
 from .api import MasterthermAPI
 from .const import (
     CHAR_MAP,
-    DEVICE_DATA_MAP,
-    DEVICE_DATA_HCMAP,
     DEVICE_INFO_MAP,
     HC_MAP,
 )
+from .datamapread import DEVICE_READ_MAP, DEVICE_READ_HCMAP
+from .datamapwrite import DEVICE_WRITE_MAP
+from .exceptions import MasterthermEntryNotFound, MasterthermPumpError
+from .special import Special
+from .__version__ import __version__
 
 _LOGGER: logging.Logger = logging.getLogger(__package__)
 
@@ -46,7 +49,6 @@ class MasterthermController:
         self.__api = MasterthermAPI(
             username, password, session, api_version=api_version
         )
-        self.__device_map = DEVICE_DATA_MAP
         self.__api_connected = False
         self.__info_update_minutes = 30
         self.__data_update_seconds = 60
@@ -66,43 +68,47 @@ class MasterthermController:
         # }
         self.__devices = {}
 
+    def __convert_data(self, item_type, item_value, registers) -> any:
+        """Convert the data for use with Populate Data."""
+        return_value: any = None
+
+        if isinstance(item_type, Special):
+            # Handle Special Types
+            special_item: Special = item_type
+
+            if special_item.condition == Special.FIXED:
+                return_value = special_item.data_type(item_value)
+            elif special_item.condition == Special.NAMEARRAY:
+                item_str = ""
+                for list_value in item_value:
+                    item_str = item_str + CHAR_MAP[int(registers[list_value])]
+                return_value = item_str
+            elif special_item.condition == Special.FORMULA:
+                values = []
+                for item in item_value[1]:
+                    # Bool needs to converted as an int not string
+                    if item[0] == bool:
+                        values.append(item[0](int(registers[item[1]])))
+                    else:
+                        values.append(item[0](registers[item[1]]))
+
+                return_value = special_item.evaluate(item_value[0], values)
+        else:
+            # Convert Simple Types
+            # Bool needs to converted as an int not string
+            if item_type == bool:
+                return_value = item_type(int(registers[item_value]))
+            else:
+                return_value = item_type(registers[item_value])
+
+        return return_value
+
     def __populate_data(self, device_map, registers) -> dict:
         """Populate the Data from the api_full_data and DeviceMap."""
         data = {}
         for key, item in device_map.items():
             if not isinstance(item, dict):
-                item_type = item[0]
-                item_value = item[1]
-                if item_type == "fixed":
-                    data[key] = item_value
-                elif item_type == "bool":
-                    if item_value == "":
-                        data[key] = False
-                    else:
-                        data[key] = registers[item_value] == "1"
-                elif item_type == "not bool":
-                    if item_value == "":
-                        data[key] = True
-                    else:
-                        data[key] = not registers[item_value] == "1"
-                elif item_type == "float":
-                    if item_value == "":
-                        data[key] = 0.0
-                    else:
-                        data[key] = float(registers[item_value])
-                elif item_type == "int":
-                    if item_value == "":
-                        data[key] = 0
-                    else:
-                        data[key] = int(registers[item_value])
-                elif item_type == "string":
-                    if item_value == "":
-                        data[key] = ""
-                    else:
-                        item_str = ""
-                        for list_value in item_value:
-                            item_str = item_str + CHAR_MAP[int(registers[list_value])]
-                        data[key] = item_str
+                data[key] = self.__convert_data(item[0], item[1], registers)
             else:
                 data[key] = self.__populate_data(device_map[key], registers)
 
@@ -110,7 +116,7 @@ class MasterthermController:
 
     def __get_hc_name(self, hc_id: int, device_key: str) -> str:
         """Build the Heating Cooling Circuit Name from the full data."""
-        if HC_MAP[hc_id]["id"] not in DEVICE_DATA_HCMAP:
+        if HC_MAP[hc_id]["id"] not in DEVICE_READ_HCMAP:
             return "0"
 
         hc_name = ""
@@ -118,7 +124,7 @@ class MasterthermController:
         full_data = self.__devices[device_key]["api_full_data"]
 
         # Get the Name from the api data.
-        for key in DEVICE_DATA_HCMAP[HC_MAP[hc_id]["id"]]["name"][1]:
+        for key in DEVICE_READ_HCMAP[HC_MAP[hc_id]["id"]]["name"][1]:
             hc_name = hc_name + CHAR_MAP[int(full_data[key])]
             hc_empty = hc_empty + "-"
 
@@ -167,15 +173,16 @@ class MasterthermController:
 
     async def __get_hp_updates(self, full_load: bool = False) -> None:
         """Refresh data and information for all the devices, apply default restrictions to
-               the number of times a call can be made..
+        the number of times a call can be made..
 
-               Parameters:
-                   full_load: Optional Force a full load of the data, defaultis false
-        gd
-               Raises
-                   MasterthermConnectionError - Failed to Connect
-                   MasterthermAuthenticationError - Failed to Authenticate
-                   MasterthermUnsupportedRole - Role is not in supported roles"""
+        Parameters:
+            full_load: Optional Force a full load of the data, defaultis false
+
+        Raises
+            MasterthermConnectionError - Failed to Connect
+            MasterthermAuthenticationError - Failed to Authenticate
+            MasterthermUnsupportedRole - Role is not in supported roles
+            MasterthermServerTimeoutError - Server Timed Out more than once."""
         for device in self.__devices.values():
             module_id = device["info"]["module_id"]
             unit_id = device["info"]["unit_id"]
@@ -199,27 +206,37 @@ class MasterthermController:
             # Refresh the Device Data, refresh rate of this data is restricted by default
             # to try and keep frequency of requests down.all(iterable)
             last_data_update = None
+            last_update_time = "0"
             if "last_data_update" in device and not full_load:
                 last_data_update = device["last_data_update"]
+                last_update_time = device["last_update_time"]
 
             if (
                 last_data_update is None
                 or datetime.now()
                 >= last_data_update + timedelta(seconds=self.__data_update_seconds)
             ):
-                # Refresh Device Data.
-                device_data = await self.__api.get_device_data(
-                    module_id, unit_id, last_update_time=device["last_update_time"]
-                )
-                device["last_data_update"] = datetime.now()
+                try:
+                    # Refresh Device Data.
+                    device_data = await self.__api.get_device_data(
+                        module_id, unit_id, last_update_time=last_update_time
+                    )
+                    device["last_data_update"] = datetime.now()
+                    device["data"]["operating_mode"] = "online"
 
-                # Check that we have data, sometimes nothing is returned.
-                if device_data["data"]:
-                    device["last_update_time"] = device_data["timestamp"]
-                    device["api_update_data"] = device_data["data"]["varData"][
-                        str(unit_id).zfill(3)
-                    ].copy()
-                    device["api_full_data"].update(device["api_update_data"])
+                    # Check that we have data, sometimes nothing is returned.
+                    if device_data["data"]:
+                        device["last_update_time"] = device_data["timestamp"]
+                        device["api_update_data"] = device_data["data"]["varData"][
+                            str(unit_id).zfill(3)
+                        ].copy()
+                        device["api_full_data"].update(device["api_update_data"])
+
+                except MasterthermPumpError as mpe:
+                    if mpe.status == MasterthermPumpError.OFFLINE:
+                        device["data"]["operating_mode"] = "offline"
+                    else:
+                        raise MasterthermPumpError(mpe.status, mpe.message) from mpe
 
     async def connect(self, reload_modules: bool = False) -> bool:
         """Connect to the API, check the supported roles and update if required.
@@ -233,7 +250,8 @@ class MasterthermController:
         Raises:
             MasterthermConnectionError - Failed to Connect
             MasterthermAuthenticationError - Failed to Authenticate
-            MasterthermUnsportedRole - Role is not supported by API"""
+            MasterthermUnsportedRole - Role is not supported by API
+            MasterthermServerTimeoutError - Server Timed Out more than once."""
         result = await self.__api.connect()
 
         # Initialize the Dictionary.
@@ -292,7 +310,8 @@ class MasterthermController:
             MasterthermConnectionError - Failed to Connect
             MasterthermNotReady - Failed to complete first full import
             MasterthermAuthenticationError - Failed to Authenticate
-            MasterthermUnsupportedRole - Role is not in supported roles"""
+            MasterthermUnsupportedRole - Role is not in supported roles
+            MasterthermServerTimeoutError - Server Timed Out more than once."""
         if not self.__api_connected:
             return False
 
@@ -311,79 +330,76 @@ class MasterthermController:
 
             # Add some additional details to the info such as URL
             device["info"]["api_url"] = self.__api.get_url()
+            device["info"]["version"] = __version__
 
-            # Populate Device Data
-            device["data"] = self.__populate_data(
-                self.__device_map, device["api_full_data"]
-            )
-            data = device["data"]
+            if not device["data"]["operating_mode"] == "offline":
+                # Populate Device Data
+                device["data"] = self.__populate_data(
+                    DEVICE_READ_MAP, device["api_full_data"]
+                )
 
-            # Update Season Setting
-            season_info = data["season_info"]
-            data["season"] = ("" if season_info["hp_seasonset"] else "auto:") + (
-                "winter" if season_info["hp_season"] else "summer"
-            )
+                # Check the Pad Names, if blank get them from the data
+                # Populate all data correctly.
+                for hc_id in range(0, 7):
+                    hc_key = HC_MAP[hc_id]["id"]
+                    hc_pad = HC_MAP[hc_id]["pad"]
 
-            # Update Operating Mode
-            operating_mode = "heating"
-            if data["domestic_hot_water"]["heating"]:
-                operating_mode = "dhw"
-            elif data["heating_circuits"]["pool"]["heating"]:
-                operating_mode = "pool"
-            else:
-                error_info = data["error_info"]
-                if (
-                    error_info["some_error"]
-                    or error_info["three_errors"]
-                    or not data["aux_heater_1"]
-                ):
-                    if data["cooling_mode"]:
-                        if data["dewp_control"]:
-                            operating_mode = "dpc"
-                        else:
-                            operating_mode = "cooling"
+                    if device["api_info"][hc_pad] in ("", "0"):
+                        device["info"][hc_pad] = self.__get_hc_name(hc_id, device_id)
 
-            device["data"]["operating_mode"] = operating_mode
+                    if device["info"][hc_pad] in ("", "0"):
+                        device["info"][hc_pad] = HC_MAP[hc_id]["default"]
 
-            # Check the Pad Names, if blank get them from the data
-            # Populate all data correctly.
-            for hc_id in range(0, 7):
-                hc_key = HC_MAP[hc_id]["id"]
-                hc_pad = HC_MAP[hc_id]["pad"]
+                    device["data"]["heating_circuits"][hc_key]["name"] = device["info"][
+                        hc_pad
+                    ]
 
-                if device["api_info"][hc_pad] in ("", "0"):
-                    device["info"][hc_pad] = self.__get_hc_name(hc_id, device_id)
+                # Disable certain fields if Cooling Mode is disabled
+                cooling_disabled = not device["info"]["cooling"] == "1"
+                if cooling_disabled:
+                    device["data"].pop("hp_function")
+                    device["data"].pop("control_curve_cooling")
 
-                if device["info"][hc_pad] in ("", "0"):
-                    device["info"][hc_pad] = HC_MAP[hc_id]["default"]
-
-                device["data"]["heating_circuits"][hc_key]["name"] = device["info"][
-                    hc_pad
+                # Set if circuits are enabled or not, remove circuts that
+                # are not enabled.
+                hc_enabled_result = self.__hc_enabled(device_id)
+                hc_circuits: dict = self.__devices[device_id]["data"][
+                    "heating_circuits"
                 ]
-
-            # Set if circuits are enabled or not, remove circuts that
-            # are not enabled.
-            hc_enabled_result = self.__hc_enabled(device_id)
-            hc_circuits: dict = self.__devices[device_id]["data"]["heating_circuits"]
-            for hc_id, hc_enabled in hc_enabled_result.items():
-                if not hc_enabled:
-                    hc_circuits.pop(hc_id)
-                else:
-                    hc_circuits[hc_id]["enabled"] = hc_enabled
-                    if hc_circuits[hc_id]["pad"]["enabled"]:
-                        hc_circuits[hc_id].pop("int")
+                for hc_id, hc_enabled in hc_enabled_result.items():
+                    if not hc_enabled:
+                        hc_circuits.pop(hc_id)
                     else:
-                        hc_circuits[hc_id].pop("pad")
+                        hc_circuits[hc_id]["enabled"] = hc_enabled
+                        if not hc_circuits[hc_id]["pad"]["enabled"]:
+                            hc_circuits[hc_id].pop("pad")
 
-            # Check if the Pool and Solar are enabled
-            if not hc_circuits["solar"]["enabled"]:
-                hc_circuits.pop("solar")
-            if not hc_circuits["pool"]["enabled"]:
-                hc_circuits.pop("pool")
+                        if (
+                            cooling_disabled
+                            and "control_curve_cooling" in hc_circuits[hc_id]
+                        ):
+                            hc_circuits[hc_id].pop("control_curve_cooling")
 
-            # Remove the Domestic Hot Water if the feature is disabled.
-            if not self.__devices[device_id]["data"]["domestic_hot_water"]["enabled"]:
-                self.__devices[device_id]["data"].pop("domestic_hot_water")
+                # Check if the Pool and Solar are enabled
+                if not hc_circuits["solar"]["enabled"]:
+                    hc_circuits.pop("solar")
+                if not hc_circuits["pool"]["enabled"]:
+                    hc_circuits.pop("pool")
+
+                # Remove the Domestic Hot Water if the feature is disabled.
+                if not self.__devices[device_id]["data"]["domestic_hot_water"][
+                    "enabled"
+                ]:
+                    self.__devices[device_id]["data"].pop("domestic_hot_water")
+
+                # Filter the Fan/ Brine and Water Pumps based on HP Type
+                # 0=A/W, 1=B/W, 2=W/W, 3=DX/W, 4=A/W R, 5=B/W R, 6=W/W R
+                data = self.__devices[device_id]["data"]
+                hp_type = data["hp_type"]
+                if not hp_type in [0, 4]:
+                    data.pop("fan_running")
+                if not hp_type in [1, 2, 3, 5, 6]:
+                    data.pop("brine_pump_running")
 
         return True
 
@@ -448,3 +464,167 @@ class MasterthermController:
             data = self.__devices[key]["data"]
 
         return data
+
+    def get_device_data_item(self, module_id: str, unit_id: str, entry: str) -> any:
+        """Get the Device Data Item based on the dot notation.
+        Parameters:
+            module_id (str): The id of the module
+            unit_id (str):   The id of the unit
+            entry (str):     The entry to using dot notation, e.g. hp_power_state
+                             or heating_circuits.hc0.on
+
+        Returns:
+            any: Returns the value of the item found
+
+        Raises:
+            MasterthermEntryNotFound - Entry is not valid."""
+        data = self.get_device_data(module_id, unit_id)
+
+        keys: list[str] = entry.split(".")
+        for i in range(len(keys) - 1):
+            if keys[i] in data:
+                data = data[keys[i]]
+            else:
+                raise MasterthermEntryNotFound(
+                    "401", f"{keys[i]} in {entry} not found."
+                )
+
+        item = keys[len(keys) - 1]
+        if not item in data:
+            raise MasterthermEntryNotFound("401", f"{item} in {entry} not found.")
+
+        return data[item]
+
+    def get_diagnostics_data(self, hide_sensitive: bool = True) -> dict:
+        """Get full diagnostics data hiding sensitive information by default.
+        Paramaeters:
+            shide_sensitive (bool): Default True, if False then sensitive information is shown.
+
+        Returns
+            dict: Returns a dict of all the data for all modules."""
+        diagnostics_data = {}
+        new_module_id = 1111
+        old_module_id = ""
+
+        for device_id, device_item in self.__devices.items():
+            module_id = device_item["info"]["module_id"]
+            unit_id = device_item["info"]["unit_id"]
+
+            if module_id != old_module_id:
+                old_module_id = module_id
+                new_module_id = new_module_id + 1
+
+            info = device_item["info"].copy()
+            api_info = device_item["api_info"].copy()
+
+            if hide_sensitive:
+                device_id = f"{str(new_module_id)}_{unit_id}"
+                info["module_id"] = str(new_module_id)
+                info["module_name"] = "REDACTED"
+                info["name"] = "REDACTED"
+                info["surname"] = "REDACTED"
+                info["latitude"] = "REDACTED"
+                info["longitude"] = "REDACTED"
+                info["place"] = "REDACTED"
+                info["notes"] = "REDACTED"
+
+                api_info["moduleid"] = str(new_module_id)
+                api_info["module_name"] = "REDACTED"
+                api_info["givenname"] = "REDACTED"
+                api_info["surname"] = "REDACTED"
+                api_info["password8"] = "REDACTED"
+                api_info["password9"] = "REDACTED"
+                api_info["password10"] = "REDACTED"
+                api_info["city"] = "REDACTED"
+                api_info["notes"] = "REDACTED"
+
+                # Version 2 API Information.
+                if "location" in api_info:
+                    api_info["location"] = "REDACTED"
+                    api_info["latitude"] = "REDACTED"
+                    api_info["longitude"] = "REDACTED"
+
+            diagnostics_data[device_id] = {
+                "info": info,
+                "data": device_item["data"].copy(),
+                "api_info": api_info,
+                "api_data": device_item["api_full_data"].copy(),
+            }
+
+        return diagnostics_data
+
+    async def set_device_data_item(
+        self, module_id: str, unit_id: str, entry: str, value: any
+    ) -> bool:
+        """Set the Device Data and Update to the Mastertherm Heat Pump.
+        Parameters:
+            module_id (str): The id of the module
+            unit_id (str):   The id of the unit
+            entry (str):     The entry to using dot notation, e.g. hp_power_state
+                             or heating_circuits.hc0.on
+            value:           The value to set should be bool, str, float or int
+
+        Returns:
+            bool: True if success, False if failure
+
+        Raises:
+            MasterthermConnectionError - Failed to Connect
+            MasterthermAuthenticationError - Failed to Authenticate
+            MasterthermEntryNotFound - Entry is not valid.
+            MasterthermServerTimeoutError - Server Timed Out more than once."""
+        # Split the entry into its components and find the mapping and data type.
+        # Check if in both read and write map, if not in both stop.
+        write_map = DEVICE_WRITE_MAP
+        device_data = self.get_device_data(module_id, unit_id)
+        keys: list[str] = entry.split(".")
+        for i in range(len(keys) - 1):
+            if keys[i] in write_map and keys[i] in device_data:
+                write_map = write_map[keys[i]]
+                device_data = device_data[keys[i]]
+            else:
+                raise MasterthermEntryNotFound(
+                    "401", f"{keys[i]} in {entry} not found."
+                )
+
+        # Make sure the item is valid the start processing.
+        item = keys[len(keys) - 1]
+        if not item in write_map or not item in device_data:
+            raise MasterthermEntryNotFound("401", f"{item} in {entry} not found.")
+
+        # Get the Registry to set and prepare the value.
+        entry_type = write_map[item][0]
+        entry_value = write_map[item][1]
+        entry_reg = ""
+
+        if isinstance(entry_type, Special):
+            # If Special then find the Registry Entry.
+            special_item: Special = entry_type
+            registers = self.get_device_registers(module_id, unit_id)
+            if special_item.condition == Special.FORMULA:
+                values = []
+                for item in entry_value[1]:
+                    if item[0] == bool:
+                        values.append(item[0](int(registers[item[1]])))
+                    else:
+                        values.append(item[0](registers[item[1]]))
+
+                entry_reg = special_item.evaluate(entry_value[0], values)
+                entry_type = special_item.data_type
+            else:
+                return False
+        else:
+            # If Normal just set it from the Value
+            entry_reg = entry_value
+
+        # Test if the entry_type matches our value type.
+        if isinstance(value, entry_type):
+            if isinstance(value, bool):
+                entry_value = str(int(value))
+            else:
+                entry_value = str(value)
+        else:
+            return False
+
+        return await self.__api.set_device_data(
+            module_id, unit_id, entry_reg, entry_value
+        )
